needs "hashek.ml";;
needs "eval/utils.ml";;

let lstring_of_string = String.map (fun c -> if c = '\n' then ' ' else c)
let lstring_of_term = lstring_of_string o string_of_term
let lstring_of_type = lstring_of_string o string_of_type

module Meson_thf = struct

let reset_terms,int_of_term,term_of_int =
  let tstore = ref ([]:(term * int)list)
  and tcounter = ref 1 in
  let reset_terms() = tcounter := 1; tstore := [] in
  let int_of_term t =
    try assoc t !tstore with Failure _ ->
    let n = !tcounter in
    incr tcounter; tstore := (t,n)::!tstore; n in
  let term_of_int i = rev_assoc i (!tstore) in
  reset_terms,int_of_term,term_of_int

let reset_types,int_of_type,type_of_int =
  let tstore = ref ([]:(hol_type * int)list)
  and tcounter = ref 1 in
  let reset_types() = tcounter := 1; tstore := [] in
  let int_of_type t =
    try assoc t !tstore with Failure _ ->
    let n = !tcounter in
    incr tcounter; tstore := (t,n)::!tstore; n in
  let type_of_int i = rev_assoc i (!tstore) in
  reset_types,int_of_type,type_of_int


let rec fold_cs_vs fn tm sofar =
  try let l,r = try dest_forall tm with Failure _ ->
                try dest_exists tm with Failure _ ->
                    dest_abs tm in
      fold_cs_vs fn r (fold_cs_vs fn l sofar)
  with Failure _ -> try
      let l,r = try dest_conj tm with Failure _ ->
                try dest_disj tm with Failure _ ->
                try dest_eq tm with Failure _ ->
                    dest_imp tm in
      fold_cs_vs fn r (fold_cs_vs fn l sofar)
  with Failure _ -> try
      let tm = dest_neg tm in fold_cs_vs fn tm sofar
  with Failure _ -> try
      let l, r = dest_comb tm in
      fold_cs_vs fn r (fold_cs_vs fn l sofar)
  with Failure _ -> fn tm sofar

let rec consts_in_type ty =
  try
    let l, r = dest_fun_ty ty in
    union (consts_in_type l) (consts_in_type r)
  with _ -> if is_vartype ty || ty = `:bool` then [] else [ty]

let collect_stuff tm (tys, cs) =
  union tys (consts_in_type (type_of tm)),
  union cs (if is_var tm || tm = `T` || tm = `F` then [] else [tm])

let tptp_of_const tm =
  if tm = `T` then "$true"
  else if tm = `F` then "$false"
  else "c" ^ string_of_int (int_of_term tm)
let tptp_of_var tm = "V" ^ string_of_int (int_of_term tm)
let rec tptp_of_type ty =
  if ty = `:bool` then "$o"
  else try
    let l, r = dest_fun_ty ty in
    "(" ^ tptp_of_type l ^ " > " ^ tptp_of_type r ^ ")"
  with _ -> "ty" ^ string_of_int (int_of_type ty)

let tptp_of_var_ty tm = tptp_of_var tm ^ " : " ^ tptp_of_type (type_of tm)

let tptp_of_form =
  let parens s = "(" ^ s ^ ")" in
  let rec of_form (env, consts as acc) tm =
    if is_neg tm then "~" ^ parens (of_form acc (dest_neg tm))
    else if is_forall tm then of_quant "!" acc (dest_forall tm)
    else if is_exists tm then of_quant "?" acc (dest_exists tm)
    else if is_abs    tm then of_quant "^" acc (dest_abs    tm)
    else if is_conj tm then of_binary " & " acc (dest_conj tm)
    else if is_disj tm then of_binary " | " acc (dest_disj tm)
    else if is_iff  tm then of_binary "<=>" acc (dest_iff  tm)
    else if is_imp  tm then of_binary " =>" acc (dest_imp  tm)
    else if is_eq   tm then of_binary " = " acc (dest_eq   tm)
    else if is_comb tm then
      let hop, args = strip_comb tm in
      parens (String.concat " @ " (List.map (of_form acc) (hop :: args)))
    else if is_var tm && not (mem tm consts) then tptp_of_var tm
    else tptp_of_const tm
  and of_quant quant (env, consts) (v, bod) =
    parens (quant ^ "[" ^ tptp_of_var_ty v ^ "]: " ^
            of_form (v::env, subtract consts [v]) bod)
  and of_binary bin acc (l, r) =
    parens (of_form acc l ^ bin ^ of_form acc r)
  in of_form

let fprint_problem ths (lths, hyp) oc =
  reset_terms(); reset_types();

  let assms = map concl ths @ map (concl o snd) lths in
  let all = hyp :: assms in
  let tmvars = freesl all in
  let tyvars = unions (map type_vars_in_term all) in
  let (tys, consts) = List.fold_right (fold_cs_vs collect_stuff) all (tyvars, tmvars) in

  List.iteri
    (fun i ty ->
      Printf.fprintf oc "%% Type: %s\n" (lstring_of_type ty);
      Printf.fprintf oc "thf(type%d, type, %s : $tType).\n" i (tptp_of_type ty)
    ) tys;

  List.iteri
    (fun i tm ->
      Printf.fprintf oc "%% Constant: %s\n" (lstring_of_term tm);
      Printf.fprintf oc "thf(const%d, type, %s : %s).\n" i (tptp_of_const tm) (tptp_of_type (type_of tm))
    ) consts;

  List.iteri
    (fun i tm ->
      Printf.fprintf oc "%% Axiom: %s\n" (lstring_of_term tm);
      Printf.fprintf oc "thf(ax%d, axiom, %s).\n" i (tptp_of_form ([], frees tm) tm)
    ) assms;

  Printf.fprintf oc "%% Conjecture: %s\n" (lstring_of_term hyp);
  Printf.fprintf oc "thf(conj, conjecture, %s).\n" (tptp_of_form ([], frees hyp) hyp)

end


module Meson_fof = struct

let tptp_of_const tm = "c" ^ string_of_int (Meson.fol_of_const tm)
let tptp_of_var   tm = "V" ^ string_of_int (Meson.fol_of_var   tm)

let parlist = function
    [] -> ""
  | xs -> "(" ^ String.concat ", " xs ^ ")"

let rec tptp_of_fol_term = function
    Fvar v -> "V" ^ string_of_int v
  | Fnapp (f, args) ->
      "c" ^ string_of_int f ^ parlist (map tptp_of_fol_term args)

let tptp_of_fol_atom (p, args) =
  "p" ^ string_of_int p ^ parlist (map tptp_of_fol_term args)

let tptp_of_form =
  let parens s = "(" ^ s ^ ")" in
  let rec of_form (env, consts as acc) tm =
    if is_neg tm then "~" ^ parens (of_form acc (dest_neg tm))
    else if is_forall tm then of_quant "!" acc (dest_forall tm)
    else if is_exists tm then of_quant "?" acc (dest_exists tm)
    else if is_conj tm then of_binary " & " acc (dest_conj tm)
    else if is_disj tm then of_binary " | " acc (dest_disj tm)
    else if is_eq   tm then
      let l, r = dest_eq tm in
      let fl = Meson.fol_of_term env consts l
      and fr = Meson.fol_of_term env consts r in
      tptp_of_fol_term fl ^ " = " ^ tptp_of_fol_term fr
    else tptp_of_fol_atom (Meson.fol_of_atom env consts tm)
  and of_quant quant (env, consts) (v, bod) =
    parens (quant ^ "[" ^ tptp_of_var v ^ "]: " ^
            of_form (v::env, subtract consts [v]) bod)
  and of_binary bin acc (l, r) =
    parens (of_form acc l ^ bin ^ of_form acc r)
  in of_form

let tptp_of_thm th =
  let lconsts = freesl (hyp th) in
  let tm = concl th in
  tptp_of_form ([], lconsts) tm

let fprint_problem ths oc = try
  Meson.reset_consts ();
  let ths = List.concat (map CONJUNCTS ths) in
  List.iter (Printf.fprintf oc "%% %s\n%!" o lstring_of_term o concl) ths;
  let hashek, axs = List.partition (fun th -> concl th = Hashek.hashek_tm) ths in
  List.iteri (fun i -> Printf.fprintf oc "fof(ax%d, axiom, %s).\n%!" i o tptp_of_thm) axs;
  assert (List.length hashek < 2);
  if List.length hashek > 0 then
    tptp_of_form ([], []) Hashek.hashek_tm |>
    Printf.fprintf oc "fof(hashek, negated_conjecture, %s).\n%!"
  with Failure f -> Printf.fprintf oc "Failure: %s\n%!" f; raise (Failure f)

end


module Meson_tptp = struct

let with_out fname f =
  let oc = open_out fname in f oc; close_out oc

let write_fol_problem dir n ths =
  let fname = Printf.sprintf "%s/mesonf%05i.p" dir n in
  Format.printf "Writing problem %s\n%!" fname;
  with_out fname (Meson_fof.fprint_problem ths)

let write_hol_problem dir n ths g =
  let fname = Printf.sprintf "%s/mesonh%05i.p" dir n in
  Format.printf "Writing problem %s\n%!" fname;
  with_out fname (Meson_thf.fprint_problem ths g)

let WRITE_FOL_TAC dir n g = write_fol_problem dir n (map snd (fst g)); ALL_TAC g

let WRITE_PREPARE_TAC ths =
  (* mostly from Meson *)
  (* in case that the goal is `F` at this point,
     REFUTE_THEN does not change the goal state
     most notably, it does not introduce any new label *)
  REFUTE_THEN (LABEL_TAC "copconcl") THEN
  Meson.POLY_ASSUME_TAC (map GEN_ALL ths) THEN

  (* in case the original goal was `F`, continue without introducing hashek *)
  (Hashek.MARK_CONCL "copconcl" ORELSE ALL_TAC) THEN

  (* unmodified Meson parts start here *)
  W(MAP_EVERY(UNDISCH_TAC o concl o snd) o fst) THEN
  SELECT_ELIM_TAC THEN
  W(fun (asl,w) -> MAP_EVERY (fun v -> SPEC_TAC(v,v)) (frees w)) THEN
  CONV_TAC(PRESIMP_CONV THENC
           TOP_DEPTH_CONV BETA_CONV THENC
           LAMBDA_ELIM_CONV THENC
           CONDS_CELIM_CONV THENC
           Meson.QUANT_BOOL_CONV) THEN
  REPEAT(GEN_TAC ORELSE DISCH_TAC) THEN
  REFUTE_THEN ASSUME_TAC THEN
  RULE_ASSUM_TAC(CONV_RULE(NNF_CONV THENC SKOLEM_CONV)) THEN
  REPEAT (FIRST_X_ASSUM CHOOSE_TAC) THEN
  ASM_FOL_TAC

let probno = ref 0

let hook ths g =
  incr probno;
  try
    let outdir = (try Sys.getenv "EVAL" with _ -> "HH") ^ "/meson-tptp" in
    mkdir_p outdir 0o755;
    write_hol_problem outdir !probno ths g;
    ignore ((WRITE_PREPARE_TAC ths THEN WRITE_FOL_TAC outdir !probno) g)
  with Failure f -> Format.printf "Write error: %s\n%!" f

end
