module Ocaml_typing = struct

type dummy

(* ------------------------------------------------------------------------- *)
(* Basic data structures copied from OCaml. May be version-dependent.        *)
(* ------------------------------------------------------------------------- *)

type label = int

(*** from typing/ident.ml: ***)

module Ident = struct

type t = { stamp: int; name: string; mutable flags: int }

type 'a tbl =
    Empty
  | Node of 'a tbl * 'a data * 'a tbl * int

and 'a data =
  { ident: t;
    data: 'a;
    previous: 'a data option }

end

(*** from typing/path.ml: ***)

module Path = struct

type t =
    Pident of Ident.t
  | Pdot of t * string * int
  | Papply of t * t

end

(*** from typing/types.ml: ***)

module Types = struct

type type_expr =
  { mutable desc: type_desc;
    mutable level: dummy;
    mutable id: dummy }

and type_desc =
    Tvar of string option
  | Tarrow of label * type_expr * type_expr * dummy
  | Ttuple of type_expr list
  | Tconstr of Path.t * type_expr list * dummy
  | Tobject of type_expr * (Path.t * type_expr list) option ref
  | Tfield of string * dummy * type_expr * type_expr
  | Tnil
  | Tlink of type_expr
  | Tsubst of type_expr         (* for copying *)
  | Tvariant of dummy
  | Tunivar of string option
  | Tpoly of type_expr * type_expr list
  | Tpackage of Path.t * dummy * type_expr list

type value_description =
  { val_type: type_expr;                (* Type of the value *)
    val_kind: dummy;
    val_loc: dummy;
    val_attributes: dummy
 }

type module_type =
    Tmty_ident of dummy
  | Tmty_signature of signature
  | Tmty_functor of dummy * dummy * dummy
  | Tmty_alias of dummy

and signature = signature_item list

and signature_item =
    Tsig_value of Ident.t * value_description
  | Tsig_type of dummy * dummy * dummy
  | Tsig_typext of dummy * dummy * dummy
  | Tsig_module of Ident.t * module_declaration * dummy
  | Tsig_modtype of dummy * dummy
  | Tsig_class of dummy * dummy * dummy
  | Tsig_class_type of dummy * dummy * dummy

and module_declaration =
  {
    md_type: module_type;
    md_attributes: dummy;
    md_loc: dummy;
  }

end


(*** from typing/env.ml: ***)

module Env = struct

open Types

module Envtbl =
  struct
    (* A table indexed by identifier, with an extra slot to record usage. *)
    type 'a t = ('a * bool ref) Ident.tbl
end;;

type env_t = {
  values: (dummy * value_description) Envtbl.t;
  constrs: dummy;
  labels: dummy;
  types: dummy;
  modules: (dummy * module_declaration) Envtbl.t;
  modtypes: dummy;
  components: dummy;
  classes: dummy;
  cltypes: dummy;
  functor_args: dummy;
  summary: dummy;
  local_constraints: dummy;
  gadt_instances: dummy;
  flags: dummy
}

end

(* ------------------------------------------------------------------------- *)
(* End of basic data structures copied from OCaml.                           *)
(* ------------------------------------------------------------------------- *)

open Ident
open Types
open Env
open Path

(* Iterate over the entries of a table. *)

let rec iter_tbl (f : Ident.t -> 'a -> unit) = function
  | Empty -> ()
  | Node (t1,d,t2,_) ->
      f d.ident d.data;
      iter_tbl f t1;
      iter_tbl f t2;;

(* If the given type is simple return its name, otherwise None. *)

let rec get_simple_type = function
  | Tlink { desc = Tconstr (Pident p,[],_) } -> Some p.name
  | Tlink { desc = d } -> get_simple_type d
  | _ -> None;;

let iter_values (p : Ident.t -> bool) (f : string -> string option -> unit) env =
  let rec iter_val pfx i vd =
    if p i then f (pfx ^ i.name) (get_simple_type vd.val_type.desc)
  and iter_mod pfx i md =
    if p i then match md.md_type with
        Tmty_signature sg ->
          let pfx' = pfx ^ i.name ^ "." in
          List.iter (function
            | Tsig_value  (i',vd)   -> iter_val pfx' i' vd
            | Tsig_module (i',md,_) -> iter_mod pfx' i' md
            | _ -> ()) sg
      | _ -> ()

  in
    iter_tbl (fun i ((_,vd),_) -> iter_val "" i vd) env.values;
    iter_tbl (fun i ((_,md),_) -> iter_mod "" i md) env.modules

end
